"""
The `kcl_error` file mainly contains all KCL exceptions.

KCLException is the top-level exception class, all KCL exceptions inherit from KCLException.

KCLException.err_info_stack :
A file information stack is built in KCLException to save file information when an exception occurs.

KCLException.gen_err_msg :
Method `gen_err_msg` generates KCL error message without highlighting
by calling the method provided in kcl_err_template.py

KCLException.show_msg_with_theme :
Method `show_msg_with_theme` generates KCL error message with highlighting
by calling the method provided in kcl_err_template.py

:note: At present, the KCL error message templates of the two methods `gen_err_msg` and `show_msg_with_theme`
are created by default arguments, and they are not synchronized.
If the template of one method is changed,
the template of the other method also needs to be replaced manually,
otherwise the message output by the two methods will be different.

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""
import sys
import typing
from enum import Enum, unique

import kclvm.kcl.error.kcl_err_template as err_template

from kclvm.kcl.error.kcl_err_msg import KCL_ERR_MSG, ErrEwcode
from kclvm.internal.util import PreCheck, CheckRules, PostCheck


class KCLException(Exception):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "ewcode")
    @PreCheck(
        (
            lambda v: CheckRules.check_str_len_allow_none(
                v, len(KCL_ERR_MSG.get_defaule_ewcode())
            )
        ),
        "ewcode",
    )
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        ewcode: str = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLException_Ew if not ewcode else ewcode
        self.name = KCL_ERR_MSG.get_err_name_by_ewcode(self.ewcode)
        self.err_info_stack: typing.List[err_template.ErrFileMsg] = []
        if file_msgs:
            self.err_info_stack.extend(file_msgs)
        self.arg_msg = (
            arg_msg
            if arg_msg
            else KCL_ERR_MSG.get_defaule_arg_msg_by_errid(self.ewcode)
        )

    def __str__(self):
        return self.gen_err_msg()

    def show_msg_with_theme(
        self,
        color_template=err_template.KCLErrMsgTemplateDefault(
            fmt=err_template.ErrMsgFmt.COLOR_TXT
        ),
    ) -> str:
        if not color_template:
            color_template = err_template.KCLErrMsgTemplateDefault(
                fmt=err_template.ErrMsgFmt.COLOR_TXT
            )
        return self.gen_err_msg(msg_tem=color_template)

    def gen_err_msg(self, msg_tem=err_template.KCLErrMsgTemplateDefault()) -> str:
        result = ""
        result += err_template.get_err_msg(self.ewcode, msg_tem)
        count = 0
        for file in reversed(self.err_info_stack):
            if file and file.filename:
                file.indent_count = count
                result += err_template.get_hint_msg(file, msg_tem)
                count += 1
        return result + "\n" + self.arg_msg if self.arg_msg else result

    def no_err_msg(self) -> bool:
        return self.err_info_stack is None or len(self.err_info_stack) == 0

    def append_err_info(self, einfo: err_template.ErrFileMsg):
        self.err_info_stack.append(einfo)

    def pop_err_info(self) -> err_template.ErrFileMsg:
        return (
            self.err_info_stack.pop(-1)
            if len(self.err_info_stack) > 0
            else err_template.ErrFileMsg()
        )

    @property
    def filename(self):
        return self.err_info_stack[-1].filename if len(self.err_info_stack) > 0 else ""

    @property
    def lineno(self):
        return self.err_info_stack[-1].line_no if len(self.err_info_stack) > 0 else ""

    @property
    def colno(self):
        return self.err_info_stack[-1].col_no if len(self.err_info_stack) > 0 else ""


class KCLError(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLWarning(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLWarning_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLSyntaxException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLSyntaxException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLCompileException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLCompileException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLRuntimeException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLRuntimeException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLAttributeException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLAttributeException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLSchemaException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLSchemaException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLMixinException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLMixinException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLInheritException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLInheritException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLImportException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLImportException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLTypeException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLTypeException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLDecoratorException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLDecoratorException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLArgumentException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLArgumentException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLOverflowException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLOverflowException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLComplingException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLComplingException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLRunningException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLRunningException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLDeprecatedException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLDeprecatedException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLDocException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLDocException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLImmutableException(KCLException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLImmutableException_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class InvalidSyntaxError(KCLError, KCLSyntaxException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.InvalidSyntax_Ew
        self.accepts_lark = []
        self.accepts_msg = []
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLTabError(KCLError, KCLSyntaxException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLTabError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLIndentationError(KCLError, KCLSyntaxException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLIndentationError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class CannotFindModule(KCLError, KCLCompileException, KCLImportException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.CannotFindModule_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class FailedLoadModule(KCLError, KCLCompileException, KCLImportException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.FailedLoadModule_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class RecursiveLoad(KCLError, KCLCompileException, KCLImportException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.RecursiveLoad_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class FloatOverflow(KCLError, KCLRuntimeException, KCLOverflowException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.FloatOverflow_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class FloatUnderflow(KCLWarning, KCLCompileException, KCLOverflowException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.FloatUnderflow_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class IntOverflow(KCLError, KCLRuntimeException, KCLOverflowException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.IntOverflow_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class InvalidDocstring(KCLWarning, KCLCompileException, KCLDocException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.InvalidDocstring_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class DeprecatedError(KCLError, KCLRuntimeException, KCLDeprecatedException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.Deprecated_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class DeprecatedWarning(KCLWarning, KCLDeprecatedException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.DeprecatedWarning_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class UnKnownDecoratorError(KCLError, KCLCompileException, KCLDecoratorException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.UnKnownDecorator_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class InvalidDecoratorTargetError(KCLError, KCLCompileException, KCLDecoratorException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.InvalidDecoratorTarget_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class MixinNamingError(KCLError, KCLCompileException, KCLMixinException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.MixinNamingError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class MixinStructureIllegal(KCLError, KCLCompileException, KCLMixinException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.MixinStructureIllegal_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class SchemaCheckFailure(KCLError, KCLRuntimeException, KCLSchemaException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.SchemaCheckFailure_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class CannotAddMembersComplieError(KCLError, KCLCompileException, KCLSchemaException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.CannotAddMembersComplieError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class CannotAddMembersRuntimeError(KCLError, KCLRuntimeException, KCLSchemaException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.CannotAddMembersRuntimeError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class IndexSignatureError(KCLError, KCLCompileException, KCLSchemaException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.IndexSignatureError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class TypeRuntimeError(KCLError, KCLRuntimeException, KCLTypeException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.TypeRuntimeError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class TypeComplieError(KCLError, KCLCompileException, KCLTypeException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.TypeComplieError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class CompileError(KCLError, KCLCompileException, KCLComplingException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.CompileError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class SelectorError(KCLError, KCLCompileException, KCLComplingException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.SelectorError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLNameError(KCLError, KCLCompileException, KCLComplingException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLNameError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLValueError(KCLError, KCLCompileException, KCLComplingException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLValueError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLKeyError(KCLError, KCLCompileException, KCLComplingException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLKeyError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class UniqueKeyError(KCLError, KCLCompileException, KCLComplingException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.UniqueKeyError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLAttributeComplieError(KCLError, KCLCompileException, KCLAttributeException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLAttributeComplieError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLAttributeRuntimeError(KCLError, KCLRuntimeException, KCLAttributeException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLAttributeRuntimeError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class IllegalAttributeError(KCLError, KCLCompileException, KCLAttributeException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.IllegalAttributeError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class MultiInheritError(KCLError, KCLCompileException, KCLInheritException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.MultiInheritError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class CycleInheritError(KCLError, KCLCompileException, KCLInheritException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.CycleInheritError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class IllegalInheritError(KCLError, KCLCompileException, KCLInheritException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.IllegalInheritError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class IllegalArgumentRuntimeError(KCLError, KCLRuntimeException, KCLArgumentException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.IllegalArgumentRuntimeError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class IllegalArgumentComplieError(KCLError, KCLCompileException, KCLArgumentException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.IllegalArgumentComplieError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class IllegalArgumentSyntaxError(KCLError, KCLSyntaxException, KCLArgumentException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.IllegalArgumentSyntaxError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class EvaluationError(KCLError, KCLRuntimeException, KCLRunningException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.EvaluationError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class InvalidFormatSpec(KCLError, KCLRuntimeException, KCLRunningException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.InvalidFormatSpec_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLAssertionError(KCLError, KCLRuntimeException, KCLRunningException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLAssertionError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class ImmutableRuntimeError(KCLError, KCLCompileException, KCLImmutableException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.ImmutableRuntimeError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class ImmutableCompileError(KCLError, KCLCompileException, KCLImmutableException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.ImmutableCompileError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class KCLRecursionError(KCLError, KCLRuntimeException, KCLRunningException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.KCLRecursionError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


class PlanError(KCLError, KCLRuntimeException, KCLRunningException):
    @PreCheck(
        (
            lambda v: CheckRules.check_list_item_type_allow_none(
                v, err_template.ErrFileMsg
            )
        ),
        "file_msgs",
    )
    @PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
    def __init__(
        self,
        file_msgs: typing.List[err_template.ErrFileMsg] = None,
        arg_msg: str = None,
    ):
        self.ewcode = ErrEwcode.PlanError_Ew
        KCLException.__init__(
            self, file_msgs=file_msgs, ewcode=self.ewcode, arg_msg=arg_msg
        )


@unique
class ErrType(Enum):
    InvalidSyntax_TYPE = (0,)
    TabError_TYPE = (1,)
    IndentationError_TYPE = (2,)
    CannotFindModule_TYPE = (3,)
    FailedLoadModule_TYPE = (4,)
    CompileError_TYPE = (5,)
    EvaluationError_TYPE = (6,)
    RecursiveLoad_TYPE = (7,)
    FloatOverflow_TYPE = (8,)
    FloatUnderflow_TYPE = (9,)
    IntOverflow_TYPE = (10,)
    InvalidDocstring_TYPE = (11,)
    Deprecated_TYPE = (12,)
    UnKnownDecorator_TYPE = (13,)
    InvalidDecoratorTarget_TYPE = (14,)
    InvalidFormatSpec_TYPE = (15,)
    SelectorError_TYPE = (16,)
    SchemaCheckFailure_TYPE = (17,)
    MixinNamingError_TYPE = (18,)
    MixinStructureIllegal_TYPE = (19,)

    IndexSignatureError_TYPE = (20,)
    TypeError_Runtime_TYPE = (21,)
    TypeError_Compile_TYPE = (22,)
    NameError_TYPE = (23,)
    ValueError_TYPE = (24,)
    KeyError_TYPE = (25,)
    UniqueKeyError_TYPE = (26,)
    AttributeError_TYPE = (27,)
    AttributeError_Runtime_TYPE = (28,)
    AssertionError_TYPE = (29,)
    ImmutableCompileError_TYPE = (30,)
    ImmutableRuntimeError_TYPE = (31,)
    MultiInheritError_TYPE = (32,)
    CycleInheritError_TYPE = (33,)
    IllegalInheritError_TYPE = (34,)
    IllegalAttributeError_TYPE = (35,)
    IllegalArgumentError_TYPE = (36,)
    IllegalArgumentError_Complie_TYPE = (37,)
    IllegalArgumentError_Syntax_TYPE = (38,)
    RecursionError_TYPE = (39,)
    PlanError_TYPE = (40,)
    Deprecated_Warning_TYPE = (41,)
    CannotAddMembers_TYPE = (42,)
    CannotAddMembers_Runtime_TYPE = (43,)


ERR_TYPE_EWCODE_MAP = {
    ErrType.InvalidSyntax_TYPE: InvalidSyntaxError,
    ErrType.TabError_TYPE: KCLTabError,
    ErrType.IndentationError_TYPE: KCLIndentationError,
    ErrType.CannotFindModule_TYPE: CannotFindModule,
    ErrType.FailedLoadModule_TYPE: FailedLoadModule,
    ErrType.CompileError_TYPE: CompileError,
    ErrType.EvaluationError_TYPE: EvaluationError,
    ErrType.RecursiveLoad_TYPE: RecursiveLoad,
    ErrType.FloatOverflow_TYPE: FloatOverflow,
    ErrType.FloatUnderflow_TYPE: FloatUnderflow,
    ErrType.IntOverflow_TYPE: IntOverflow,
    ErrType.InvalidDocstring_TYPE: InvalidDocstring,
    ErrType.Deprecated_TYPE: DeprecatedError,
    ErrType.Deprecated_Warning_TYPE: DeprecatedWarning,
    ErrType.UnKnownDecorator_TYPE: UnKnownDecoratorError,
    ErrType.InvalidDecoratorTarget_TYPE: InvalidDecoratorTargetError,
    ErrType.InvalidFormatSpec_TYPE: InvalidFormatSpec,
    ErrType.SelectorError_TYPE: SelectorError,
    ErrType.SchemaCheckFailure_TYPE: SchemaCheckFailure,
    ErrType.MixinNamingError_TYPE: MixinNamingError,
    ErrType.MixinStructureIllegal_TYPE: MixinStructureIllegal,
    ErrType.CannotAddMembers_TYPE: CannotAddMembersComplieError,
    ErrType.CannotAddMembers_Runtime_TYPE: CannotAddMembersRuntimeError,
    ErrType.IndexSignatureError_TYPE: IndexSignatureError,
    ErrType.TypeError_Runtime_TYPE: TypeRuntimeError,
    ErrType.TypeError_Compile_TYPE: TypeComplieError,
    ErrType.NameError_TYPE: KCLNameError,
    ErrType.ValueError_TYPE: KCLValueError,
    ErrType.KeyError_TYPE: KCLKeyError,
    ErrType.UniqueKeyError_TYPE: UniqueKeyError,
    ErrType.AttributeError_TYPE: KCLAttributeComplieError,
    ErrType.AttributeError_Runtime_TYPE: KCLAttributeRuntimeError,
    ErrType.AssertionError_TYPE: KCLAssertionError,
    ErrType.ImmutableCompileError_TYPE: ImmutableCompileError,
    ErrType.ImmutableRuntimeError_TYPE: ImmutableRuntimeError,
    ErrType.MultiInheritError_TYPE: MultiInheritError,
    ErrType.CycleInheritError_TYPE: CycleInheritError,
    ErrType.IllegalInheritError_TYPE: IllegalInheritError,
    ErrType.IllegalAttributeError_TYPE: IllegalAttributeError,
    ErrType.IllegalArgumentError_TYPE: IllegalArgumentRuntimeError,
    ErrType.IllegalArgumentError_Complie_TYPE: IllegalArgumentComplieError,
    ErrType.IllegalArgumentError_Syntax_TYPE: IllegalArgumentSyntaxError,
    ErrType.RecursionError_TYPE: KCLRecursionError,
    ErrType.PlanError_TYPE: PlanError,
}


@PreCheck(
    (lambda v: CheckRules.check_list_item_type_allow_none(v, err_template.ErrFileMsg)),
    "file_msgs",
)
@PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrType)), "err_type")
@PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
@PostCheck(lambda result: CheckRules.check_type_not_none(result, KCLException))
def get_exception(
    err_type: ErrType = None,
    file_msgs: typing.List[err_template.ErrFileMsg] = None,
    arg_msg: str = None,
):
    return ERR_TYPE_EWCODE_MAP[err_type](file_msgs=file_msgs, arg_msg=arg_msg)


@PreCheck(
    (lambda v: CheckRules.check_list_item_type_allow_none(v, err_template.ErrFileMsg)),
    "file_msgs",
)
@PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
@PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrType)), "err_type")
def report_exception(
    err_type: ErrType,
    file_msgs: typing.List[err_template.ErrFileMsg] = None,
    arg_msg: str = None,
):
    raise get_exception(err_type, file_msgs, arg_msg)


@PreCheck(
    (lambda v: CheckRules.check_list_item_type_allow_none(v, err_template.ErrFileMsg)),
    "file_msgs",
)
@PreCheck((lambda v: CheckRules.check_type_allow_none(v, str)), "arg_msg")
@PreCheck((lambda v: CheckRules.check_type_not_none(v, ErrType)), "err_type")
def report_warning(
    err_type: ErrType,
    file_msgs: typing.List[err_template.ErrFileMsg] = None,
    arg_msg: str = None,
):
    print_kcl_warning_message(
        get_exception(err_type, file_msgs, arg_msg), file=sys.stderr
    )


@PreCheck((lambda v: CheckRules.check_type_not_none(v, KCLException)), "err")
def print_kcl_error_message(err: KCLException, file=sys.stderr):
    err_msg = err.show_msg_with_theme() if file.isatty() else (str(err))
    print(err_msg, file=file)


@PreCheck((lambda v: CheckRules.check_type_not_none(v, KCLWarning)), "err")
def print_kcl_warning_message(err: KCLWarning, file=sys.stderr):
    err_msg = err.show_msg_with_theme() if file.isatty() else (str(err))
    print(err_msg, file=file)


@PreCheck((lambda v: CheckRules.check_type_not_none(v, Exception)), "err")
def print_common_error_message(err: Exception, file=sys.stderr):
    print("Error: {0}".format(err), file=file)


@PreCheck((lambda v: CheckRules.check_type_allow_none(v, Exception)), "err")
def print_internal_error_message(err: Exception = None, file=sys.stderr):
    if err:
        print(err, file=file)
    print("Internal Error! Please report a bug to us.", file=file)



# Copyright 2020 The KCL Authors. All rights reserved.

from typing import Callable, cast

import kclvm.api.object as objpkg

from .type import Type


def WalkType(tpe: Type, walk_fn: Callable):
    """Walk one type recursively and deal the type using the `walk_fn`"""
    if not tpe or not isinstance(tpe, Type):
        return tpe
    tpe = walk_fn(tpe)
    if tpe.type_kind() == objpkg.KCLTypeKind.UnionKind:
        tpe = cast(objpkg.KCLUnionTypeObject, tpe)
        types = []
        for t in tpe.types:
            t = WalkType(t, walk_fn)
            if t:
                types.append(t)
        tpe.types = types
    elif tpe.type_kind() == objpkg.KCLTypeKind.ListKind:
        tpe = cast(objpkg.KCLListTypeObject, tpe)
        tpe.item_type = WalkType(tpe.item_type, walk_fn)
    elif tpe.type_kind() == objpkg.KCLTypeKind.DictKind:
        tpe = cast(objpkg.KCLDictTypeObject, tpe)
        tpe.key_type = WalkType(tpe.key_type, walk_fn)
        tpe.value_type = WalkType(tpe.value_type, walk_fn)
    return tpe



from typing import List

import kclvm.kcl.ast.ast as ast
import kclvm.tools.lint.message.message as message
from kclvm.tools.lint.checkers.base_checker import BaseChecker

MSGS = {
    "E0501": (
        "line too long (%d > %d characters).",
        "Line too long.",
        "line too long ('{0}' > '{1}' characters).",
    )
}


class MiscChecker(BaseChecker):
    def __init__(self, linter) -> None:
        super().__init__(linter)
        self.name = "MiscCheck"
        self.code = None
        self.MSGS = MSGS
        self.module = None
        self.prog = None
        self.work_dir = None
        self.root: str = None

    def reset(self) -> None:
        self.msgs.clear()
        self.code = None
        self.module = None
        self.root = None

    def get_module(self, prog: ast.Program, code: str) -> None:
        assert isinstance(prog, ast.Program)
        assert code is not None
        self.prog = prog
        self.module = prog.pkgs["__main__"][0]
        self.code = code
        self.root: str = prog.root

    def check(self, prog: ast.Program, code: str) -> None:
        assert isinstance(prog, ast.Program)
        assert code is not None
        self.reset()
        self.get_module(prog, code)
        code_lines = self.code.split("\n")
        self.check_line_too_long(code_lines)

    # todo: check in ast instead of code or file
    def check_line_too_long(self, code_lines: List[str]) -> None:
        for i, v in enumerate(code_lines):
            if len(v) > self.options.max_line_length:
                self.msgs.append(
                    message.Message(
                        "E0501",
                        self.module.filename,
                        MSGS["E0501"][0]
                        % (
                            len(v),
                            self.options.max_line_length,
                        ),
                        v.strip(),
                        (i + 1, 1),
                        [len(v), self.options.max_line_length],
                    )
                )



#! /usr/bin/env python3
import ipaddress as _ip
import socket as _socket

import kclvm.kcl.error as kcl_error


def check_empty_str(ip):
    if not isinstance(ip, str) or ip.strip() == "":
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.EvaluationError_TYPE,
            arg_msg="ip must be non-empty string",
        )


def _get_ip(ip):
    try:
        return _ip.ip_address(ip)
    except ValueError:
        return None


def KMANGLED_split_host_port(ip_end_point: str):
    """
    split the 'host' and 'port' from the ip end point
    """
    check_empty_str(ip_end_point)
    return ip_end_point.split(":")


def KMANGLED_join_host_port(host, port):
    """
    merge the 'host' and 'port'
    """
    return "{}:{}".format(host, port)


def KMANGLED_fqdn(name=""):
    """
    get Fully Qualified Domain Name (FQDN)
    """
    return _socket.getfqdn(str(name))


def KMANGLED_parse_IP(ip):
    """
    parse 'ip' to a real IP address
    """
    return _get_ip(ip)


def KMANGLED_to_IP4(ip):
    """
    get the IP4 form of 'ip'
    """
    return str(_get_ip(ip))


def KMANGLED_to_IP16(ip):
    """
    get the IP16 form of 'ip'
    """
    return int(_get_ip(ip))


def KMANGLED_IP_string(ip: str):
    """
    get the IP string
    """
    return _get_ip(ip)


def KMANGLED_is_IPv4(ip: str):
    """
    whether 'ip' is a IPv4 one
    """
    ip = _get_ip(ip)
    return isinstance(ip, _ip.IPv4Address)


def KMANGLED_is_IP(ip: str) -> bool:
    """
    whether ip is a valid ip address

    Parameters
    ----------
    - ip: input ip address

    Returns
    -------
    - is_ip: a bool type return value
    """
    ip = _get_ip(ip)
    return ip is not None


def KMANGLED_is_loopback_IP(ip: str):
    """
    whether 'ip' is a loopback one
    """
    ip = _get_ip(ip)
    return ip.is_loopback if ip else False


def KMANGLED_is_multicast_IP(ip: str):
    """
    whether 'ip' is a multicast one
    """
    ip = _get_ip(ip)
    return ip.is_multicast if ip else False


def KMANGLED_is_interface_local_multicast_IP(ip: str):
    """
    whether 'ip' is a interface, local and multicast one
    """
    try:
        ip = _ip.ip_interface(ip)
        return (ip.is_site_local and ip.is_multicast) if ip else False
    except ValueError:
        return False


def KMANGLED_is_link_local_multicast_IP(ip: str):
    """
    whether 'ip' is a link local and multicast one
    """
    ip = _get_ip(ip)
    return (ip.is_link_local and ip.is_multicast) if ip else False


def KMANGLED_is_link_local_unicast_IP(ip: str):
    """
    whether 'ip' is a link local and unicast one
    """
    ip = _get_ip(ip)
    return (ip.is_link_local and not ip.is_multicast) if ip else False


def KMANGLED_is_global_unicast_IP(ip: str):
    """
    whether 'ip' is a global and unicast one
    """
    ip = _get_ip(ip)
    return (ip.is_global and not ip.is_multicast) if ip else False


def KMANGLED_is_unspecified_IP(ip: str):
    """
    whether 'ip' is a unspecified one
    """
    ip = _get_ip(ip)
    return ip.is_unspecified if ip else False



#! /usr/bin/env python3

import kclvm.config


def write_out(inputs):
    outputs = inputs
    if kclvm.config.output:
        with open(kclvm.config.output, "w") as f:
            f.write(outputs)
    else:
        print(outputs, end="")


LOG_INDENT_STRING = "    "
log_indent = 0


def write_log(message, level=1):
    """Write log message whose level is no less than the verbosity level"""
    if kclvm.config.verbose >= level:
        for _ in range(log_indent):
            print(LOG_INDENT_STRING, end="")
        print(message)


def indent_log(step=1):
    """Adjust the indentation level of log messages"""
    global log_indent
    log_indent += step



#! /usr/bin/env python3

import os
import json
from typing import Union, List, Dict, Optional
from copy import deepcopy

from lark.exceptions import UnexpectedCharacters, UnexpectedToken
from lark.tree import Tree as LarkTree
from lark.lexer import Token as LarkToken

import kclvm.kcl.error as kcl_error
from kclvm.compiler.parser.lark_parser import GetKclLarkParser

TREE_TYPE = "tree"
TOKEN_TYPE = "token"


class Token:
    ASSIGN = "ASSIGN"  # "="
    COLON = "COLON"  # ":"
    SEMI_COLON = "SEMI_COLON"  # ";"
    COMMA = "COMMA"  # ","
    LEFT_PARENTHESES = "LEFT_PARENTHESES"  # "("
    RIGHT_PARENTHESES = "RIGHT_PARENTHESES"  # ")"
    LEFT_BRACKETS = "LEFT_BRACKETS"  # "["
    RIGHT_BRACKETS = "RIGHT_BRACKETS"  # "]"
    LEFT_BRACE = "LEFT_BRACE"  # "{"
    RIGHT_BRACE = "RIGHT_BRACE"  # "}"
    PLUS = "PLUS"  # "+"
    MINUS = "MINUS"  # "-"
    MULTIPLY = "MULTIPLY"  # "*"
    DIVIDE = "DIVIDE"  # "/"
    MOD = "MOD"  # "%"
    DOT = "DOT"  # "."
    AND = "AND"  # "&"
    OR = "OR"  # "|"
    XOR = "XOR"  # "^"
    NOT = "NOT"  # "~"
    LESS_THAN = "LESS_THAN"  # "<"
    GREATER_THAN = "GREATER_THAN"  # ">"
    EQUAL_TO = "EQUAL_TO"  # "=="
    NOT_EQUAL_TO = "NOT_EQUAL_TO"  # "!="
    GREATER_THAN_OR_EQUAL_TO = "GREATER_THAN_OR_EQUAL_TO"  # ">="
    LESS_THAN_OR_EQUAL_TO = "LESS_THAN_OR_EQUAL_TO"  # "<="
    DOUBLE_STAR = "DOUBLE_STAR"  # "**"
    DOUBLE_DIVIDE = "DOUBLE_DIVIDE"  # "//"
    SHIFT_LEFT = "SHIFT_LEFT"  # "<<"
    SHIFT_RIGHT = "SHIFT_RIGHT"  # ">>"

    COMP_PLUS = "COMP_PLUS"  # "+="
    COMP_MINUS = "COMP_MINUS"  # "-="
    COMP_MULTIPLY = "COMP_MULTIPLY"  # "*="
    COMP_DIVIDE = "COMP_DIVIDE"  # "/="
    COMP_MOD = "COMP_MOD"  # "%="
    COMP_AND = "COMP_AND"  # "&="
    COMP_OR = "COMP_OR"  # "|="
    COMP_XOR = "COMP_XOR"  # "^="
    COMP_NOT = "COMP_NOT"  # "~="
    COMP_DOUBLE_STAR = "COMP_DOUBLE_STAR"  # "**="
    COMP_DOUBLE_DIVIDE = "COMP_DOUBLE_DIVIDE"  # "//="
    COMP_SHIFT_LEFT = "COMP_SHIFT_LEFT"  # "<<="
    COMP_SHIFT_RIGHT = "COMP_SHIFT_RIGHT"  # ">>="

    IMPORT = "IMPORT"  # "import"
    AS = "AS"  # "as"
    DEF = "DEF"  # "def"
    LAMBDA = "LAMBDA"  # "lambda"
    SCHEMA = "SCHEMA"  # "schema"
    MIXIN = "MIXIN"  # "mixin"
    PROTOCOL = "PROTOCOL"  # "protocol"
    RELAXED = "RELAXED"  # "relaxed"
    CHECK = "CHECK"  # "check"
    INIT = "INIT"  # "init"
    TYPE = "TYPE"  # "type"
    FOR = "FOR"  # "for"
    ASSERT = "ASSERT"  # "assert"
    IF = "IF"  # "if"
    ELIF = "ELIF"  # "elif"
    ELSE = "ELSE"  # "else"
    L_OR = "L_OR"  # "or"
    L_AND = "L_AND"  # "and"
    L_NOT = "NOT"  # "not"
    L_L_NOT = "L_NOT"

    IN = "IN"  # "in"
    IS = "IS"  # "is"
    FINAL = "FINAL"  # "final"

    ALL = "ALL"
    ANY = "ANY"
    MAP = "MAP"
    FILTER = "FILTER"

    TRUE = "TRUE"
    FALSE = "FALSE"
    NONE = "NONE"

    NAME = "NAME"
    COMMENT = "COMMENT"
    NEWLINE = "NEWLINE"

    STRING = "STRING"
    LONG_STRING = "LONG_STRING"

    DEC_NUMBER = "DEC_NUMBER"
    HEX_NUMBER = "HEX_NUMBER"
    OCT_NUMBER = "OCT_NUMBER"
    BIN_NUMBER = "BIN_NUMBER"
    FLOAT_NUMBER = "FLOAT_NUMBER"
    IMAG_NUMBER = "IMAG_NUMBER"

    RIGHT_ARROW = "RIGHT_ARROW"

    @staticmethod
    def is_string(token: str):
        return token in [Token.STRING, Token.LONG_STRING]


class Tree:
    DICT_KEY = "dict_key"
    DICT_COMP = "dict_comp"
    LIST_COMP = "list_comp"
    DICT_EXPR = "dict_expr"
    LIST_EXPR = "list_expr"
    MULTI_EXPR = "multi_expr"
    DOUBLE_STAR_EXPR = "double_star_expr"
    CONFIG_EXPR = "config_expr"
    SUB_SCRIPT = "subscript"
    IDENTIFIER = "identifier"
    STR_CALL_EXPR = "str_call_expr"
    CALL_EXPR = "call_expr"
    CALL_SUFFIX = "call_suffix"
    SLICE_SUFFIX = "slice_suffix"
    COMP_ITER = "comp_iter"
    COMP_FOR = "comp_for"
    COMP_IF = "comp_if"
    COMP_CLAUSE = "comp_clause"
    TEST = "test"
    OR_TEST = "or_test"
    SIMPLE_EXPR = "simple_expr"
    EXPR = "expr"
    PRIMARY_EXPR = "primary_expr"
    STMT = "stmt"
    IF_STMT = "if_stmt"
    SCHEMA_MEMBER_STMT = ("schema_member_stmt",)
    SCHEMA_BODY = "schema_body"
    SCHEMA_ARGUMENTS = "schema_arguments"
    MIXINS = ("mixins",)
    TYPE_ALIAS_STMT = "type_alias_stmt"
    SCHEMA_STMT = "schema_stmt"
    SIMPLE_STMT = "simple_stmt"
    DOC_STMT = "doc_stmt"
    COMPOUND_STMT = "compound_stmt"
    SMALL_STMT = "small_stmt"
    IMPORT_STMT = "import_stmt"
    ASSIGN_STMT = "assign_stmt"
    ASSERT_STMT = "assert_stmt"
    EXPR_STMT = "expr_stmt"
    MEMBER_STMT = "member_stmt"
    TYPE = "type"
    EXECUTION_BLOCK = "execution_block"
    STRING_DOT_NAME = "string_dot_name"

    SELECTOR_EXPR = "selector_expr"
    SELECTOR_SUFFIX = "selector_suffix"
    LIST_SELECTOR_SUFFIX = "list_selector_suffix"
    DICT_SELECTOR_SUFFIX = "dict_selector_suffix"


BRACKETS_TOKENS = [
    Token.LEFT_PARENTHESES,  # "("
    Token.RIGHT_PARENTHESES,  # ")"
    Token.LEFT_BRACKETS,  # "["
    Token.RIGHT_BRACKETS,  # "]"
    Token.LEFT_BRACE,  # "{"
    Token.RIGHT_BRACE,  # "}"
]
OPERATOR_TOKENS = {
    Token.ASSIGN,  # "="
    Token.PLUS,  # "+"
    Token.MINUS,  # "-"
    Token.MULTIPLY,  # "*"
    Token.DIVIDE,  # "/"
    Token.MOD,  # "%"
    Token.AND,  # "&"
    Token.OR,  # "|"
    Token.XOR,  # "^"
    Token.NOT,  # "~"
    Token.LESS_THAN,  # "<"
    Token.GREATER_THAN,  # ">"
    Token.EQUAL_TO,  # "=="
    Token.NOT_EQUAL_TO,  # "!="
    Token.GREATER_THAN_OR_EQUAL_TO,  # ">="
    Token.LESS_THAN_OR_EQUAL_TO,  # "<="
    Token.DOUBLE_STAR,  # "**"
    Token.DOUBLE_DIVIDE,  # "//"
    Token.SHIFT_LEFT,  # "<<"
    Token.SHIFT_RIGHT,  # ">>"
    Token.COMP_PLUS,  # "+="
    Token.COMP_MINUS,  # "-="
    Token.COMP_MULTIPLY,  # "*="
    Token.COMP_DIVIDE,  # "/="
    Token.COMP_MOD,  # "%="
    Token.COMP_AND,  # "&="
    Token.COMP_OR,  # "|="
    Token.COMP_XOR,  # "^="
    Token.COMP_NOT,  # "~="
    Token.COMP_DOUBLE_STAR,  # "**="
    Token.COMP_DOUBLE_DIVIDE,  # "//="
    Token.COMP_SHIFT_LEFT,  # "<<="
    Token.COMP_SHIFT_RIGHT,  # ">>="
}
SEPERATOR_TOKENS = {
    Token.COLON,  # ":"
    Token.SEMI_COLON,  # ";"
    Token.COMMA,  # ","
}
FUNCTION_EXPRS = {
    Tree.CALL_EXPR,
    Tree.STR_CALL_EXPR,
}
_WALK_FUNCTION_PREFIX = "walk_"

AstType = Dict[str, Union[str, List]]


class TreeWalker:
    """
    The TreeWalk class can be used as a superclass in order
    to walk an AST or similar tree.

    This class is meant to be subclassed, with the subclass adding walker
    methods.

    Per default the walker functions for the nodes are ``'walk_'`` +
    class name of the node.  So a `expr_stmt` node visit function would
    be `walk_expr_stmt`.  This behavior can be changed by overriding
    the `walk` method.  If no walker function exists for a node
    (return value `None`) the `generic_walker` walker is used instead.
    """

    def __init__(self) -> None:
        pass

    def walk(self, node: AstType) -> None:
        """Visit a node."""
        name = node["name"]
        method = _WALK_FUNCTION_PREFIX + name
        walker = getattr(self, method, self.generic_walk)
        walker(node)

    def generic_walk(self, node: AstType):
        """Called if no explicit walker function exists for a node."""
        children_key = "children"
        if children_key in node:
            for n in node[children_key]:
                self.walk(n)
        else:
            self.walk(node)

    def walk_nodes(self, *nodes: Union[AstType, str]) -> None:
        """Write nodes"""
        if not nodes:
            return
        for node in nodes:
            self.walk_node(node)

    def walk_node(self, node: Union[AstType, str]) -> None:
        """Write node"""
        self.walk(node)

    def get(self, node: AstType, name: str) -> Optional[AstType]:
        """
        Get children from 'node' named 'name'
        """
        if not node or "children" not in node:
            return None
        for i, n in enumerate(node["children"]):
            if n["name"] == name:
                node["children"] = node["children"][i + 1 :]
                return n
        return None

    def has(self, node: AstType, name: str) -> bool:
        """
        Whether 'node' has the children named 'name'
        """
        return name in [n["name"] for n in node["children"]]

    def get_value(self, node: AstType, default: Optional[str] = None) -> Optional[str]:
        """
        Get tree node value recursively
        """
        if not node:
            return default
        if node["type"] == TOKEN_TYPE:
            return node.get("value", default)
        elif node["type"] == TREE_TYPE:
            return "".join(self.get_value(n) for n in node["children"])
        return default

    def get_children_value(
        self, node: AstType, name: Optional[str] = None, default: Optional[str] = None
    ) -> Optional[str]:
        """
        Get children value recursively from 'node' named 'name'
        """
        return self.get_value(self.get(node, name), default) if name else default

    def get_internal(
        self, node: AstType, begin_name: str, end_name: str
    ) -> Optional[List[AstType]]:
        """
        Get children from 'begin_name' to 'end_name'
        """
        begin_index = -1
        end_index = -1
        for i, n in enumerate(node["children"]):
            if n["name"] == begin_name and begin_index == -1:
                begin_index = i
            if n["name"] == end_name:
                end_index = i
        if end_index >= begin_index:
            result = deepcopy(node["children"][begin_index : end_index + 1])
            node["children"] = node["children"][end_index + 1 :]
            return result
        return None


def build_children(t: LarkTree) -> List[AstType]:
    """
    Build json ast children of tree node, children type can be token or tree
    """
    return list(map(lambda c: walk_lark_tree(c), t.children))


def build_token_content(t: LarkToken) -> AstType:
    """
    Build json ast token node
    """
    return {
        "type": TOKEN_TYPE,
        "name": t.type,
        "value": t.value,
        "line": t.line,
        "column": t.column,
    }


def build_tree_content(t: LarkTree) -> AstType:
    """
    Build json ast tree node
    """
    return {
        "type": TREE_TYPE,
        "name": t.data,
        "children": build_children(t),
    }


def walk_lark_tree(t: Union[LarkTree, LarkToken]) -> AstType:
    """
    Traverse the lark tree and return python dict
    """
    if isinstance(t, LarkTree):
        return build_tree_content(t)
    elif isinstance(t, LarkToken):
        return build_token_content(t)
    else:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CompileError_TYPE,
            arg_msg="Unknown node type: {}".format(type(t)),
        )


def get_lark_tree_from_expr(
    source: str, to_json_str: bool = True
) -> Union[str, AstType]:
    try:
        # Get lark parse tree
        parse_tree = GetKclLarkParser().parse(source + "\n")
        # Convert python dict to standard json
        ast = walk_lark_tree(parse_tree)
        return (json.dumps(ast) + "\n") if to_json_str else ast
    except (
        UnexpectedCharacters,
        UnexpectedToken,
    ) as err:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.InvalidSyntax_TYPE,
            file_msgs=[kcl_error.ErrFileMsg(line_no=err.line, col_no=err.column)],
        )
    except Exception as err:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CompileError_TYPE, arg_msg=str(err)
        )


def get_lark_tree_from_file(filename: str, to_json_str: bool = True) -> str:
    """
    Get kcl json ast from .k file
    """
    with open(filename, "r", encoding="utf-8") as pyfile:
        source = pyfile.read()
    return get_lark_tree_from_expr(source, to_json_str)


def get_lark_tree_from_path(path: str) -> List[str]:
    """
    Get kcl json ast from kmodule package
    """
    return [get_lark_tree_from_file(file) for file in sorted(os.listdir(path))]



# Copyright 2021 The KCL Authors. All rights reserved.

import re
import pathlib
import pickle
import os
import hashlib
import time

from dataclasses import dataclass
from typing import Dict

import kclvm.api.version as ver_pkg
import kclvm.api.object as objpkg
import kclvm.kcl.ast as ast

from filelock import FileLock

# -------------
# Type alias
# -------------

CacheInfo = str
Cache = Dict[str, CacheInfo]

LOCK_SUFFIX = ".lock"
NORMAL_CACHE_SUFFIX = ".data"
BYTECODE_CACHE_PREFIX = "bytecode"
BYTECODE_CACHE_SUFFIX = ".kclc"
DEFAULT_CACHE_DIR = str(pathlib.Path(".kclvm").joinpath("cache"))
FST_CACHE_DIR = str(pathlib.Path(".kclvm").joinpath("fst_cache"))
CACHE_INFO_FILENAME = "info.pickle"


@dataclass
class CacheOption:
    cache_dir: str = DEFAULT_CACHE_DIR


def _get_cache_dir(root: str, cache_dir: str = DEFAULT_CACHE_DIR) -> str:
    return os.path.join(root, cache_dir, f"{ver_pkg.VERSION}-{ver_pkg.CHECKSUM}")


def _get_cache_filename(
    root: str, pkgpath: str, cache_dir: str = DEFAULT_CACHE_DIR
) -> str:
    return os.path.join(
        root,
        cache_dir,
        f"{ver_pkg.VERSION}-{ver_pkg.CHECKSUM}",
        f"{pkgpath}{NORMAL_CACHE_SUFFIX}",
    )


def _get_cache_info_filename(root: str, cache_dir: str = DEFAULT_CACHE_DIR):
    return os.path.join(
        root, cache_dir, f"{ver_pkg.VERSION}-{ver_pkg.CHECKSUM}", CACHE_INFO_FILENAME
    )


def _get_cache_bytecode_filename(
    root: str, check_sum: str, cache_dir: str = DEFAULT_CACHE_DIR
):
    return os.path.join(
        root,
        cache_dir,
        f"{ver_pkg.VERSION}-{ver_pkg.CHECKSUM}",
        f"{BYTECODE_CACHE_PREFIX}_{check_sum}{BYTECODE_CACHE_SUFFIX}",
    )


def read_info_cache(root: str, cache_dir: str = DEFAULT_CACHE_DIR) -> Cache:
    """Read the cache if it exists and is well formed.
    If it is not well formed, the call to write_info_cache later should resolve the issue.
    """
    cache_file = pathlib.Path(_get_cache_info_filename(root, cache_dir=cache_dir))
    cache_dir = _get_cache_dir(root, cache_dir=cache_dir)
    pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)
    with FileLock(str(cache_file) + LOCK_SUFFIX):
        try:
            if not cache_file.exists():
                return {}
            with cache_file.open("rb") as fobj:
                cache: Cache = pickle.load(fobj)
                return cache
        except Exception:
            return {}


def write_info_cache(
    cache: Cache, root: str, filepath: str, cache_dir: str = DEFAULT_CACHE_DIR
) -> None:
    """Update the cache info file."""
    dst_filename = _get_cache_info_filename(root, cache_dir=cache_dir)
    try:
        cache_dir = _get_cache_dir(root, cache_dir=cache_dir)
        pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)
        relative_path = filepath.replace(root, ".", 1)
        new_cache = {
            **cache,
            **{relative_path: get_cache_info(filepath)},
        }
        tmp_filename = str(
            pathlib.Path(cache_dir).joinpath(f".{os.getpid()}{time.time_ns()}.tmp")
        )
        # Write cache atomic
        with FileLock(dst_filename + LOCK_SUFFIX):
            f = open(tmp_filename, "wb")
            pickle.dump(new_cache, f)
            f.flush()
            os.fsync(f.fileno())
            f.close()
            if pathlib.Path(dst_filename).exists():
                os.remove(dst_filename)
            os.rename(tmp_filename, dst_filename)
    except Exception as err:
        raise err


def get_cache_info(path: str) -> CacheInfo:
    """Return the information used to check if a file or path is already changed or not."""
    path = pathlib.Path(path)
    check_sum = hashlib.md5()
    if os.path.isfile(path):
        with open(path, "rb") as f:
            check_sum.update(f.read())
    else:
        for file in list(sorted(path.glob("*.k"))):
            with open(file, "rb") as f:
                check_sum.update(f.read())
    return check_sum.hexdigest()


def get_pkg_realpath_from_pkgpath(root: str, pkgpath: str) -> str:
    """Get the pkgpath real path in the file system according to the root and pkgpath"""
    filepath = str(pathlib.Path(root).joinpath(f"{pkgpath.replace('.', os.sep)}"))
    if os.path.isfile(f"{filepath}.k"):
        filepath = f"{filepath}.k"
    return filepath


def load_data_from_file(filename) -> any:
    with FileLock(filename + LOCK_SUFFIX):
        f = open(filename, "rb")
        # PyCharm
        # noinspection PyBroadException
        try:
            x = pickle.load(f)
            f.close()
            return x
        except Exception:
            f.close()
            os.remove(filename)
            return None


def save_data_to_file(dst_filename: str, tmp_filename: str, x: any):
    try:
        with FileLock(dst_filename + LOCK_SUFFIX):
            # write cache atomic
            f = open(tmp_filename, "wb")
            # PyCharm
            # noinspection PyBroadException
            pickle.dump(x, f)
            f.flush()
            os.fsync(f.fileno())
            f.close()
            if pathlib.Path(dst_filename).exists():
                os.remove(dst_filename)
            os.rename(tmp_filename, dst_filename)
            return
    except Exception as err:
        raise err


def LoadPkgCache(root: str, pkgpath: str, option: CacheOption = CacheOption()) -> any:
    if not root or not pkgpath:
        return None

    filename = _get_cache_filename(root, pkgpath, cache_dir=option.cache_dir)
    if not os.path.exists(filename):
        return None

    # Compare the md5 using cache
    realpath = get_pkg_realpath_from_pkgpath(root, pkgpath)
    if os.path.exists(realpath):
        cache_info = read_info_cache(root, cache_dir=option.cache_dir)
        relative_path = realpath.replace(root, ".", 1)
        path_info_in_cache = cache_info.get(relative_path)
        path_info = get_cache_info(realpath)
        if path_info_in_cache != path_info:
            return None

    return load_data_from_file(filename)


def SavePkgCache(root: str, pkgpath: str, x: any, option: CacheOption = CacheOption()):
    if not root or not pkgpath or not x:
        return

    dst_filename = _get_cache_filename(root, pkgpath, cache_dir=option.cache_dir)

    # Save the pkgpath timesample and filesize into the cache
    realpath = get_pkg_realpath_from_pkgpath(root, pkgpath)
    if os.path.exists(realpath):
        cache_info = read_info_cache(root, cache_dir=option.cache_dir)
        write_info_cache(cache_info, root, realpath, cache_dir=option.cache_dir)

    cache_dir = _get_cache_dir(root, cache_dir=option.cache_dir)
    pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)

    tmp_filename = str(
        pathlib.Path(cache_dir).joinpath(f"{pkgpath}.{os.getpid()}{time.time_ns()}.tmp")
    )

    save_data_to_file(dst_filename, tmp_filename, x)


def LoadMainPkgCache(
    root: str, filename: str, option: CacheOption = CacheOption()
) -> any:
    if not root or not filename:
        return None

    cache_name = filename.replace(root, "").replace("/", "_").replace("\\", "_")
    cache_filename = _get_cache_filename(root, cache_name, cache_dir=option.cache_dir)

    if not os.path.exists(cache_filename):
        return None

    # Compare the md5 using cache
    if os.path.exists(filename):
        cache_info = read_info_cache(root, cache_dir=option.cache_dir)
        relative_path = filename.replace(root, ".", 1)
        path_info_in_cache = cache_info.get(relative_path)
        path_info = get_cache_info(filename)
        if path_info_in_cache != path_info:
            return None

    return load_data_from_file(cache_filename)


def SaveMainPkgCache(
    root: str, filename: str, x: any, option: CacheOption = CacheOption()
):
    if not root or not filename:
        return

    cache_name = filename.replace(root, "").replace("/", "_").replace("\\", "_")
    dst_filename = _get_cache_filename(root, cache_name, cache_dir=option.cache_dir)

    if os.path.exists(filename):
        cache_info = read_info_cache(root, cache_dir=option.cache_dir)
        write_info_cache(cache_info, root, filename, cache_dir=option.cache_dir)

    cache_dir = _get_cache_dir(root, cache_dir=option.cache_dir)
    pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)

    tmp_filename = str(
        pathlib.Path(cache_dir).joinpath(
            f"{cache_name}.{os.getpid()}{time.time_ns()}.tmp"
        )
    )

    save_data_to_file(dst_filename, tmp_filename, x)


def LoadBytecodeCache(
    root: str, ast_program: ast.Program, option: CacheOption = CacheOption()
) -> objpkg.KCLProgram:
    if not root:
        return None
    if not ast_program or not isinstance(ast_program, ast.Program):
        return None
    if not ast_program.pkgs:
        return None
    check_sum = ast_program.get_check_sum(root)
    cache_filename = _get_cache_bytecode_filename(
        root, check_sum, cache_dir=option.cache_dir
    )
    if not os.path.exists(cache_filename):
        return None
    return load_data_from_file(cache_filename)


def SaveBytecodeCache(
    root: str,
    ast_program: ast.Program,
    program: objpkg.KCLProgram,
    option: CacheOption = CacheOption(),
):
    if not root:
        return
    if not ast_program or not isinstance(ast_program, ast.Program):
        return
    if not program or not isinstance(program, objpkg.KCLProgram):
        return
    pkgs = list(program.pkgs.keys()) if program.pkgs else None
    if not pkgs:
        return
    check_sum = ast_program.get_check_sum(root)
    dst_filename = _get_cache_bytecode_filename(
        root, check_sum, cache_dir=option.cache_dir
    )
    if os.path.exists(dst_filename):
        return
    cache_dir = _get_cache_dir(root, cache_dir=option.cache_dir)
    pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)

    tmp_filename = str(
        pathlib.Path(cache_dir).joinpath(f"{os.getpid()}{time.time_ns()}.kclc.tmp")
    )

    save_data_to_file(dst_filename, tmp_filename, program)


def IsAbsPkgPath(s: str) -> bool:
    if not s or not isinstance(s, str):
        return False
    if s.startswith("."):
        return False
    if os.path.isabs(s):
        return False
    if ".." in s:
        return False
    if re.search(r"\s", s):
        return False

    return True


def IsRelPkgPath(s: str) -> bool:
    return s.strip().startswith(".") if s and isinstance(s, str) else False


def FixImportPath(root: str, filepath: str, import_path: str) -> str:
    """
    relpath: import .sub
    FixImportPath(root, "path/to/app/file.k", ".sub")        => path.to.app.sub
    FixImportPath(root, "path/to/app/file.k", "..sub")       => path.to.sub
    FixImportPath(root, "path/to/app/file.k", "...sub")      => path.sub
    FixImportPath(root, "path/to/app/file.k", "....sub")     => sub
    FixImportPath(root, "path/to/app/file.k", ".....sub")    => ""

    abspath: import path.to.sub
    FixImportPath(root, "path/to/app/file.k", "path.to.sub") => path.to.sub
    """
    assert root
    assert filepath
    assert import_path

    if not import_path.startswith("."):
        return import_path

    # Filepath to pkgpath
    pkgpath: str = (
        os.path.relpath(os.path.dirname(filepath), root).replace("/", ".").rstrip(".")
    )
    # On the window platform
    pkgpath = pkgpath.replace("\\", ".").rstrip(".")

    leading_dot_count = len(import_path)
    for i in range(len(import_path)):
        if import_path[i] != ".":
            leading_dot_count = i
            break

    # The pkgpath is the current root path
    if not pkgpath:
        return import_path.lstrip(".") if leading_dot_count <= 1 else ""

    if leading_dot_count == 1:
        return pkgpath + import_path

    ss = pkgpath.split(".")

    if (leading_dot_count - 1) < len(ss):
        return (
            ".".join(ss[: -(leading_dot_count - 1)])
            + "."
            + import_path[leading_dot_count:]
        )

    if (leading_dot_count - 1) == len(ss):
        return import_path[leading_dot_count:]

    return ""



# Copyright 2020 The KCL Authors. All rights reserved.

from typing import Union, List, Optional

import kclvm.kcl.error as kcl_error
import kclvm.kcl.ast as ast
import kclvm.api.object as objpkg

from .type import (
    Type,
    ANY_TYPE,
    INT_TYPE,
    FLOAT_TYPE,
    STR_TYPE,
    BOOL_TYPE,
    FALSE_LIT_TYPE,
    NONE_TYPE,
    NUMBER_TYPE_KINDS,
    ITERABLE_KINDS,
    STR_KINDS,
    INT_KINDS,
    BOOL_KINDS,
    BUILTIN_KINDS,
    sup,
    has_any_type,
    is_upper_bound,
    infer_to_variable_type,
    literal_union_type_to_variable_type,
    is_kind_type_or_kind_union_type,
)


ZERO_LIT_TYPES: List[Type] = [
    objpkg.KCLIntLitTypeObject(0),
    objpkg.KCLFloatLitTypeObject(0.0),
    FALSE_LIT_TYPE,
]


def binary(
    t1: Type,
    t2: Type,
    op: Union[ast.BinOp, ast.AugOp],
    filename: Optional[str] = None,
    line: Optional[int] = None,
    column: Optional[int] = None,
) -> Type:
    """Binary operator calculation table.

    Arithmetic (int or float; result has type float unless both operands have type int)
       number + number              # addition
       number - number              # subtraction
       number * number              # multiplication
       number / number              # real division  (result is always a float)
       number // number             # floored division
       number % number              # remainder of floored division
       number ^ number              # bitwise XOR
       number << number             # bitwise left shift
       number >> number             # bitwise right shift

    Concatenation
        string + string
        list + list

    Repetition (string/list)
        int * sequence
        sequence * int

    Union
        int | int
        list | list
        dict | dict
        schema | schema
        schema | dict

    Add: number + number, str + str, list + list
    Sub: number - number
    Mul: number * number, number * list, list * number
    Div: number / number
    FloorDiv: number // number
    Mod: number % number
    Pow: number ** number
    LShift: int >> int
    RShift: int << int
    BitOr: int | int, list | list, dict | dict, schema | schema, schema | dict
    BitXOr: int ^ int
    BitAdd int & int

    And: any_type and any_type -> bool
    Or: any_type1 or any_type1 -> sup([any_type1, any_type2])
    """

    def number_binary() -> Type:
        return (
            FLOAT_TYPE
            if (
                t1.type_kind()
                in [objpkg.KCLTypeKind.FloatKind, objpkg.KCLTypeKind.FloatLitKind]
                or t2.type_kind()
                in [objpkg.KCLTypeKind.FloatKind, objpkg.KCLTypeKind.FloatLitKind]
            )
            else INT_TYPE
        )

    if has_any_type([t1, t2]):
        return ANY_TYPE

    t1 = literal_union_type_to_variable_type(t1)
    t2 = literal_union_type_to_variable_type(t2)

    if op == ast.BinOp.Add or op == ast.AugOp.Add:
        if t1.type_kind() in NUMBER_TYPE_KINDS and t2.type_kind() in NUMBER_TYPE_KINDS:
            return number_binary()
        if t1.type_kind() in STR_KINDS and t2.type_kind() in STR_KINDS:
            return STR_TYPE
        if isinstance(t1, objpkg.KCLListTypeObject) and isinstance(
            t2, objpkg.KCLListTypeObject
        ):
            return objpkg.KCLListTypeObject(item_type=sup([t1.item_type, t2.item_type]))
    elif op == ast.BinOp.Mul or op == ast.AugOp.Mul:
        if t1.type_kind() in NUMBER_TYPE_KINDS and t2.type_kind() in NUMBER_TYPE_KINDS:
            return number_binary()
        if t1.type_kind() in INT_KINDS and (
            is_kind_type_or_kind_union_type(
                t2, STR_KINDS + NUMBER_TYPE_KINDS + [objpkg.KCLTypeKind.ListKind]
            )
        ):
            return t2
        if (
            is_kind_type_or_kind_union_type(
                t1, STR_KINDS + NUMBER_TYPE_KINDS + [objpkg.KCLTypeKind.ListKind]
            )
        ) and t2.type_kind() in INT_KINDS:
            return t1
    elif op == ast.BinOp.Sub or op == ast.AugOp.Sub:
        if t1.type_kind() in NUMBER_TYPE_KINDS and t2.type_kind() in NUMBER_TYPE_KINDS:
            return number_binary()
    elif op == ast.BinOp.Div or op == ast.AugOp.Div:
        if t1.type_kind() in NUMBER_TYPE_KINDS and t2.type_kind() in NUMBER_TYPE_KINDS:
            if t2 in ZERO_LIT_TYPES:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.CompileError_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=filename, line_no=line, col_no=column
                        )
                    ],
                    arg_msg="integer division or modulo by zero",
                )
            return number_binary()
    elif op == ast.BinOp.Mod or op == ast.AugOp.Mod:
        if t1.type_kind() in NUMBER_TYPE_KINDS and t2.type_kind() in NUMBER_TYPE_KINDS:
            if t2 in ZERO_LIT_TYPES:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.CompileError_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=filename, line_no=line, col_no=column
                        )
                    ],
                    arg_msg="integer division or modulo by zero",
                )
            return INT_TYPE
    elif op == ast.BinOp.FloorDiv or op == ast.AugOp.FloorDiv:
        if t1.type_kind() in NUMBER_TYPE_KINDS and t2.type_kind() in NUMBER_TYPE_KINDS:
            if t2 in ZERO_LIT_TYPES:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.CompileError_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=filename, line_no=line, col_no=column
                        )
                    ],
                    arg_msg="integer division or modulo by zero",
                )
            return number_binary()
    elif op == ast.BinOp.Pow or op == ast.AugOp.Pow:
        if t1.type_kind() in NUMBER_TYPE_KINDS and t2.type_kind() in NUMBER_TYPE_KINDS:
            return number_binary()
    elif op == ast.BinOp.BitOr or op == ast.AugOp.BitOr:
        if t1.type_kind() in INT_KINDS and t2.type_kind() in INT_KINDS:
            return INT_TYPE
        if t2 == NONE_TYPE:
            return t1
        if t1 == NONE_TYPE:
            return t2
        if isinstance(t1, objpkg.KCLListTypeObject) and isinstance(
            t2, objpkg.KCLListTypeObject
        ):
            return objpkg.KCLListTypeObject(item_type=sup([t1.item_type, t2.item_type]))
        if isinstance(t1, objpkg.KCLDictTypeObject) and isinstance(
            t2, objpkg.KCLDictTypeObject
        ):
            return objpkg.KCLDictTypeObject(
                key_type=sup([t1.key_type, t2.key_type]),
                value_type=sup([t1.value_type, t2.value_type]),
            )
        if isinstance(t1, objpkg.KCLSchemaTypeObject) and isinstance(
            t2, (objpkg.KCLSchemaTypeObject, objpkg.KCLDictTypeObject)
        ):
            return t1
    elif op == ast.BinOp.LShift or op == ast.AugOp.LShift:
        if t1.type_kind() in INT_KINDS and t2.type_kind() in INT_KINDS:
            return INT_TYPE
    elif op == ast.BinOp.RShift or op == ast.AugOp.RShift:
        if t1.type_kind() in INT_KINDS and t2.type_kind() in INT_KINDS:
            return INT_TYPE
    elif op == ast.BinOp.BitXor or op == ast.AugOp.BitXor:
        if t1.type_kind() in INT_KINDS and t2.type_kind() in INT_KINDS:
            return INT_TYPE
    elif op == ast.BinOp.BitAnd or op == ast.AugOp.BitAnd:
        if t1.type_kind() in INT_KINDS and t2.type_kind() in INT_KINDS:
            return INT_TYPE
    elif op == ast.BinOp.And:
        return BOOL_TYPE
    elif op == ast.BinOp.Or:
        return sup([t1, t2])
    elif op == ast.BinOp.As:
        if not is_upper_bound(infer_to_variable_type(t1), t2):
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE,
                file_msgs=[
                    kcl_error.ErrFileMsg(filename=filename, line_no=line, col_no=column)
                ],
                arg_msg=f"Conversion of type '{t1.type_str()}' to type '{t2.type_str()}' "
                "may be a mistake because neither type sufficiently overlaps with the other",
            )
        return t2.schema_type if isinstance(t2, objpkg.KCLSchemaDefTypeObject) else t2
    kcl_error.report_exception(
        err_type=kcl_error.ErrType.CompileError_TYPE,
        file_msgs=[
            kcl_error.ErrFileMsg(filename=filename, line_no=line, col_no=column)
        ],
        arg_msg=f"unsupported operand type(s) for {ast.OPERATOR_VALUE_MAP.get(op)}: '{t1.type_str()}' and '{t2.type_str()}'",
    )


def compare(
    t1: Type,
    t2: Type,
    op: ast.CmpOp,
    filename: Optional[str] = None,
    line: Optional[int] = None,
    column: Optional[int] = None,
) -> Type:
    """Compare operator calculation table

    bool            # False < True            False < True
    int             # mathematical            1 < 2
    float           # as defined by IEEE 754  1.0 < 2.0
    string          # lexicographical         "1" < 2
    list            # lexicographical         [1] == [2]
    iterable        # 1 in [1, 2, 3], "s" in "ss", "key" in Schema
    """

    t1 = literal_union_type_to_variable_type(t1)
    t2 = literal_union_type_to_variable_type(t2)

    if has_any_type([t1, t2]):
        return ANY_TYPE
    if (
        is_kind_type_or_kind_union_type(t1, NUMBER_TYPE_KINDS)
        and is_kind_type_or_kind_union_type(t2, NUMBER_TYPE_KINDS)
        and op not in [ast.CmpOp.In, ast.CmpOp.NotIn]
    ):
        return BOOL_TYPE
    if (
        is_kind_type_or_kind_union_type(t1, STR_KINDS)
        and is_kind_type_or_kind_union_type(t2, STR_KINDS)
        and op not in [ast.CmpOp.Eq, ast.CmpOp.NotEq]
    ):
        return BOOL_TYPE
    if (
        is_kind_type_or_kind_union_type(t1, BUILTIN_KINDS)
        and is_kind_type_or_kind_union_type(t2, BUILTIN_KINDS)
        and op in [ast.CmpOp.Eq, ast.CmpOp.NotEq]
    ):
        return BOOL_TYPE
    if isinstance(t1, objpkg.KCLListTypeObject) and isinstance(
        t2, objpkg.KCLListTypeObject
    ):
        return BOOL_TYPE
    if (
        isinstance(t1, (objpkg.KCLDictTypeObject, objpkg.KCLSchemaTypeObject))
        and isinstance(t2, (objpkg.KCLDictTypeObject, objpkg.KCLSchemaTypeObject))
        and op in [ast.CmpOp.Eq, ast.CmpOp.NotEq]
    ):
        return BOOL_TYPE
    if op in [ast.CmpOp.In, ast.CmpOp.NotIn] and t2.type_kind() in ITERABLE_KINDS:
        return BOOL_TYPE
    if (t1 == NONE_TYPE or t2 == NONE_TYPE) and op in [
        ast.CmpOp.Eq,
        ast.CmpOp.NotEq,
        ast.CmpOp.Is,
        ast.CmpOp.IsNot,
        ast.CmpOp.Not,
    ]:
        return BOOL_TYPE
    kcl_error.report_exception(
        err_type=kcl_error.ErrType.CompileError_TYPE,
        file_msgs=[
            kcl_error.ErrFileMsg(filename=filename, line_no=line, col_no=column)
        ],
        arg_msg=f"unsupported operand type(s) for {ast.OPERATOR_VALUE_MAP.get(op)}: '{t1.type_str()}' and '{t2.type_str()}'",
    )


def unary(
    t: Type,
    op: ast.UnaryOp,
    filename: Optional[str] = None,
    line: Optional[int] = None,
    column: Optional[int] = None,
) -> Type:
    """Unary operator calculation table

    + number        unary positive          (int, float)
    - number        unary negation          (int, float)
    ~ number        unary bitwise inversion (int)
    not x           logical negation        (any type)
    """
    if has_any_type([t]):
        return ANY_TYPE

    t = literal_union_type_to_variable_type(t)

    if op == ast.UnaryOp.UAdd:
        if t.type_kind() in NUMBER_TYPE_KINDS:
            return t
    if op == ast.UnaryOp.USub:
        if t.type_kind() in NUMBER_TYPE_KINDS:
            return t
    if op == ast.UnaryOp.Invert:
        if t.type_kind() in (INT_KINDS + BOOL_KINDS):
            return INT_TYPE
    if op == ast.UnaryOp.Not:
        return BOOL_TYPE
    kcl_error.report_exception(
        err_type=kcl_error.ErrType.CompileError_TYPE,
        file_msgs=[
            kcl_error.ErrFileMsg(filename=filename, line_no=line, col_no=column)
        ],
        arg_msg=f"bad operand type for unary {ast.OPERATOR_VALUE_MAP.get(op)}: '{t.type_str()}'",
    )



# Copyright 2020 The KCL Authors. All rights reserved.

import os
import typing


def KMANGLED_arguments(name: str, value: typing.Union[bool, int, float, str]) -> None:
    """Set arguments for option function in test."""

    assert isinstance(name, str)
    assert isinstance(value, (bool, int, float, str))

    # TODO: KMANGLED_arguments Support complex parameter types

    assert name, f"testing.arguments: '{name}' is invalid name"
    return


def KMANGLED_setting_file(filename: str) -> None:
    """Set setting file for option function in test."""
    assert isinstance(filename, str)

    assert os.path.exists(filename), f"testing.setting_file: '{filename}' not exists"
    assert os.path.isfile(filename), f"testing.setting_file: '{filename}' is not file"
    return



# Copyright 2021 The KCL Authors. All rights reserved.

import abc
import sys
import traceback
import typing

from dataclasses import dataclass

import kclvm.kcl.error as kcl_error
import kclvm.internal.gpyrpc.varint as varint

from google.protobuf import message as _message
import google.protobuf.json_format as _json_format

from .protorpc_wire_pb2 import (
    MAX_REQUEST_HEADER_LEN,
    RequestHeader,
    ResponseHeader,
)


@dataclass
class Status:
    error: str = ""


class Channel(object):
    def __init__(self, *, stdin=sys.stdin, stdout=sys.stdout):
        self.stdin = stdin
        self.stdout = stdout
        self.next_id = 1

    def get_next_id(self) -> int:
        next_id = self.next_id
        self.next_id = self.next_id + 1
        return next_id

    def send_frame(self, data: bytes):
        self.stdout.buffer.write(varint.encode(len(data)))
        self.stdout.buffer.write(data)
        self.stdout.flush()

    def recv_frame(self, _max_size: int = 0) -> bytes:
        size = varint.decode_stream(self.stdin.buffer)

        if size > _max_size > 0:
            raise Exception(f"protorpc: varint overflows maxSize({_max_size})")

        data = self.stdin.buffer.read(size)
        return data

    def write_request(self, method: str, req: _message.Message):
        body = req.SerializeToString()

        hdr = RequestHeader(
            id=self.get_next_id(), method=method, raw_request_len=len(body)
        )
        self.send_frame(hdr.SerializeToString())
        self.send_frame(body)

    def read_request_header(self) -> RequestHeader:
        data = self.recv_frame(MAX_REQUEST_HEADER_LEN)
        hdr = RequestHeader()
        hdr.ParseFromString(data)

        if hdr.snappy_compressed_request_len != 0:
            raise Exception("py: unsupport snappy compressed request")
        if hdr.checksum != 0:
            raise Exception("py: unsupport checksum request")

        return hdr

    def read_request_body(self, header: RequestHeader, body: _message.Message):
        data = self.recv_frame(
            max(header.raw_request_len, header.snappy_compressed_request_len)
        )
        body.ParseFromString(data)

    def write_response(self, id_: int, error: str, response: _message.Message):
        body = response.SerializeToString()
        hdr = ResponseHeader(id=id_, error=error, raw_response_len=len(body))
        self.send_frame(hdr.SerializeToString())
        self.send_frame(body)

    def read_response_header(self) -> ResponseHeader:
        data = self.recv_frame(0)
        hdr = ResponseHeader()
        hdr.ParseFromString(data)

        if hdr.snappy_compressed_response_len != 0:
            raise Exception("py: unsupport snappy compressed response")
        if hdr.checksum != 0:
            raise Exception("py: unsupport checksum response")

        return hdr

    def read_response_body(self, header: ResponseHeader, body: _message.Message):
        data = self.recv_frame(
            max(header.raw_response_len, header.snappy_compressed_response_len)
        )
        body.ParseFromString(data)

    def call_method(
        self, method: str, req: _message.Message, resp: _message.Message
    ) -> Status:
        self.write_request(method, req)
        resp_hdr = self.read_response_header()
        self.read_response_body(resp_hdr, resp)
        return Status(error=resp_hdr.error)


class ServiceMeta(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_service_name(self) -> str:
        pass

    @abc.abstractmethod
    def get_method_list(self) -> typing.List[str]:
        pass

    @abc.abstractmethod
    def create_method_req_message(self, method: str) -> _message.Message:
        pass

    @abc.abstractmethod
    def create_method_resp_message(self, method: str) -> _message.Message:
        pass

    @abc.abstractmethod
    def get_service_instance(self) -> _message.Message:
        pass

    @abc.abstractmethod
    def call_method(self, method: str, req: _message.Message) -> _message.Message:
        pass


class Server:
    def __init__(self):
        self.srv_table: typing.Dict[str, ServiceMeta] = {}
        self.chan: typing.Optional[Channel] = None

    def register_service(self, srv: ServiceMeta):
        self.srv_table[srv.get_service_name()] = srv

    def get_service_name_list(self) -> typing.List[str]:
        name_list: typing.List[str] = []
        for s in self.srv_table.keys():
            name_list.append(s)
        name_list.sort()
        return name_list

    def get_method_name_list(self) -> typing.List[str]:
        name_list: typing.List[str] = []
        for srv in self.srv_table.values():
            srv_name = srv.get_service_name()
            for method_name in srv.get_method_list():
                name_list.append(f"{srv_name}.{method_name}")
        name_list.sort()
        return name_list

    def run(self, *, stdin=sys.stdin, stdout=sys.stdout):
        self.chan = Channel(stdin=stdin, stdout=stdout)
        while True:
            self._accept_one_call()

    def run_once(self, *, stdin=sys.stdin, stdout=sys.stdout):
        self.chan = Channel(stdin=stdin, stdout=stdout)
        self._accept_one_call()

    def call_method(
        self, method_path: str, req_body: bytes, *, encoding: str = "json"
    ) -> dict:
        if encoding not in ["json", "protobuf"]:
            raise Exception(f"encoding '{encoding}' not support")

        service_name = method_path[: method_path.rfind(".")]
        method_name = method_path[method_path.rfind(".") + 1 :]

        if service_name not in self.srv_table:
            raise Exception(f"service '{service_name}' not found")

        service = self.srv_table[service_name]

        req = service.create_method_req_message(method_name)

        if encoding == "json":
            _json_format.Parse(req_body, req)
        else:
            req.ParseFromString(req_body)

        try:
            resp = service.call_method(method_name, req)

        except kcl_error.KCLException as err:
            raise err

        except OSError as err:
            err_message = f"OSError: {err}"
            raise Exception(err_message)

        except AssertionError as err:
            err_message = f"AssertionError: {err}"
            raise Exception(err_message)

        except Exception as err:
            err_message = f"Exception: Internal Error! Please report a bug to us: method={method_name}, err={err}, stack trace={traceback.format_exc()}"
            raise Exception(err_message)

        # return response
        # https://googleapis.dev/python/protobuf/latest/google/protobuf/json_format.html
        return _json_format.MessageToDict(
            resp,
            including_default_value_fields=True,
            preserving_proto_field_name=True,
        )

    def _accept_one_call(self):
        hdr = self._read_req_header()

        service_name = hdr.method[: hdr.method.rfind(".")]
        method_name = hdr.method[hdr.method.rfind(".") + 1 :]

        if service_name not in self.srv_table:
            raise Exception(f"service '{service_name}' not found")

        service = self.srv_table[service_name]

        req = self._read_req(service, hdr)

        try:
            resp = service.call_method(method_name, req)
            self._write_resp(hdr.id, "", resp)
            return

        except kcl_error.KCLException as err:
            resp = service.create_method_resp_message(method_name)

            err_message = f"{err}"
            self._write_resp(hdr.id, err_message, resp)
            return

        except OSError as err:
            resp = service.create_method_resp_message(method_name)

            err_message = f"OSError: {err}"
            self._write_resp(hdr.id, err_message, resp)
            return

        except AssertionError as err:
            resp = service.create_method_resp_message(method_name)

            err_message = f"AssertionError: {err}"
            self._write_resp(hdr.id, err_message, resp)
            return

        except Exception as err:
            resp = service.create_method_resp_message(method_name)

            err_message = f"Exception: Internal Error! Please report a bug to us: method={method_name}, err={err}, stack trace={traceback.format_exc()}"
            self._write_resp(hdr.id, err_message, resp)
            return

    def _read_req_header(self) -> RequestHeader:
        return self.chan.read_request_header()

    def _read_req(self, service: ServiceMeta, hdr: RequestHeader) -> _message.Message:
        req = service.create_method_req_message(hdr.method)
        self.chan.read_request_body(hdr, req)
        return req

    def _write_resp(self, id_: int, error: str, resp: _message.Message):
        self.chan.write_response(id_, error, resp)


class Client:
    def __init__(self, chan: Channel = None):
        self.chan: Channel = chan

    def call_method(
        self, method: str, req: _message.Message, resp: _message.Message
    ) -> Status:
        return self.chan.call_method(method, req, resp)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: model.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0bmodel.proto\x12\x0fkclvm.doc.model"\x83\x01\n\tModuleDoc\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rrelative_path\x18\x02 \x01(\t\x12\x0b\n\x03\x64oc\x18\x03 \x01(\t\x12+\n\x07schemas\x18\x04 \x03(\x0b\x32\x1a.kclvm.doc.model.SchemaDoc\x12\x17\n\x0fsource_code_url\x18\x05 \x01(\t"\x9d\x01\n\tSchemaDoc\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03\x64oc\x18\x02 \x01(\t\x12\x37\n\nattributes\x18\x03 \x03(\x0b\x32#.kclvm.doc.model.SchemaAttributeDoc\x12\x10\n\x08\x65xamples\x18\x04 \x01(\t\x12*\n\x0b\x62\x61se_schema\x18\x05 \x01(\x0b\x32\x15.kclvm.doc.model.Type"\x80\x01\n\x12SchemaAttributeDoc\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03\x64oc\x18\x02 \x01(\t\x12#\n\x04type\x18\x04 \x01(\x0b\x32\x15.kclvm.doc.model.Type\x12\x13\n\x0bis_optional\x18\x05 \x01(\x08\x12\x15\n\rdefault_value\x18\x06 \x01(\t"\x9c\x06\n\x04Type\x12\x10\n\x08type_str\x18\x01 \x01(\t\x12\x39\n\rtype_category\x18\x02 \x01(\x0e\x32".kclvm.doc.model.Type.TypeCategory\x12\x33\n\nunion_type\x18\x03 \x01(\x0b\x32\x1a.kclvm.doc.model.UnionTypeH\x00\x88\x01\x01\x12\x31\n\tlist_type\x18\x04 \x01(\x0b\x32\x19.kclvm.doc.model.ListTypeH\x01\x88\x01\x01\x12\x31\n\tdict_type\x18\x05 \x01(\x0b\x32\x19.kclvm.doc.model.DictTypeH\x02\x88\x01\x01\x12\x35\n\x0bschema_type\x18\x06 \x01(\x0b\x32\x1b.kclvm.doc.model.SchemaTypeH\x03\x88\x01\x01\x12<\n\x0c\x62uiltin_type\x18\x07 \x01(\x0e\x32!.kclvm.doc.model.Type.BuiltinTypeH\x04\x88\x01\x01\x12/\n\x08lit_type\x18\x08 \x01(\x0b\x32\x18.kclvm.doc.model.LitTypeH\x05\x88\x01\x01\x12@\n\x0enamed_constant\x18\t \x01(\x0e\x32#.kclvm.doc.model.Type.NamedConstantH\x06\x88\x01\x01"\x83\x01\n\x0cTypeCategory\x12\t\n\x05UNION\x10\x00\x12\x08\n\x04LIST\x10\x01\x12\x08\n\x04\x44ICT\x10\x02\x12\n\n\x06SCHEMA\x10\x03\x12\x0b\n\x07\x42UILTIN\x10\x04\x12\x07\n\x03LIT\x10\x05\x12\x12\n\x0eNAMED_CONSTANT\x10\x06\x12\x07\n\x03\x41NY\x10\x07\x12\x15\n\x11NUMBER_MULTIPLIER\x10\x08"7\n\x0b\x42uiltinType\x12\x07\n\x03INT\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\n\n\x06STRING\x10\x02\x12\x08\n\x04\x42OOL\x10\x03"\x19\n\rNamedConstant\x12\x08\n\x04NONE\x10\x00\x42\r\n\x0b_union_typeB\x0c\n\n_list_typeB\x0c\n\n_dict_typeB\x0e\n\x0c_schema_typeB\x0f\n\r_builtin_typeB\x0b\n\t_lit_typeB\x11\n\x0f_named_constant"1\n\tUnionType\x12$\n\x05types\x18\x01 \x03(\x0b\x32\x15.kclvm.doc.model.Type"4\n\x08ListType\x12(\n\titem_type\x18\x01 \x01(\x0b\x32\x15.kclvm.doc.model.Type"^\n\x08\x44ictType\x12\'\n\x08key_type\x18\x01 \x01(\x0b\x32\x15.kclvm.doc.model.Type\x12)\n\nvalue_type\x18\x02 \x01(\x0b\x32\x15.kclvm.doc.model.Type"1\n\nSchemaType\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rrelative_path\x18\x02 \x01(\t"\x9d\x01\n\x07LitType\x12\x17\n\nstring_lit\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x14\n\x07int_lit\x18\x02 \x01(\x03H\x01\x88\x01\x01\x12\x16\n\tfloat_lit\x18\x03 \x01(\x02H\x02\x88\x01\x01\x12\x15\n\x08\x62ool_lit\x18\x04 \x01(\x08H\x03\x88\x01\x01\x42\r\n\x0b_string_litB\n\n\x08_int_litB\x0c\n\n_float_litB\x0b\n\t_bool_litb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "model_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_MODULEDOC"]._serialized_start = 33
    _globals["_MODULEDOC"]._serialized_end = 164
    _globals["_SCHEMADOC"]._serialized_start = 167
    _globals["_SCHEMADOC"]._serialized_end = 324
    _globals["_SCHEMAATTRIBUTEDOC"]._serialized_start = 327
    _globals["_SCHEMAATTRIBUTEDOC"]._serialized_end = 455
    _globals["_TYPE"]._serialized_start = 458
    _globals["_TYPE"]._serialized_end = 1254
    _globals["_TYPE_TYPECATEGORY"]._serialized_start = 931
    _globals["_TYPE_TYPECATEGORY"]._serialized_end = 1062
    _globals["_TYPE_BUILTINTYPE"]._serialized_start = 1064
    _globals["_TYPE_BUILTINTYPE"]._serialized_end = 1119
    _globals["_TYPE_NAMEDCONSTANT"]._serialized_start = 1121
    _globals["_TYPE_NAMEDCONSTANT"]._serialized_end = 1146
    _globals["_UNIONTYPE"]._serialized_start = 1256
    _globals["_UNIONTYPE"]._serialized_end = 1305
    _globals["_LISTTYPE"]._serialized_start = 1307
    _globals["_LISTTYPE"]._serialized_end = 1359
    _globals["_DICTTYPE"]._serialized_start = 1361
    _globals["_DICTTYPE"]._serialized_end = 1455
    _globals["_SCHEMATYPE"]._serialized_start = 1457
    _globals["_SCHEMATYPE"]._serialized_end = 1506
    _globals["_LITTYPE"]._serialized_start = 1509
    _globals["_LITTYPE"]._serialized_end = 1666
# @@protoc_insertion_point(module_scope)



from abc import ABC, abstractmethod

import kclvm.tools.docs.factory as factory
import kclvm.tools.docs.formats as doc_formats
import kclvm.tools.docs.model_pb2 as model


class DocEscaper(ABC):
    """
    doc escaper checks the special character in the doc and escape it
    """

    @abstractmethod
    def escape(self, doc: model.ModuleDoc) -> model.ModuleDoc:
        pass


class MarkdownDocEscaper(DocEscaper):
    def escape(self, module: model.ModuleDoc) -> model.ModuleDoc:
        # remove line breaks at the beginning and the end
        module.doc = self.escape_special_symbol(module.doc.strip("\n"))
        for schema in module.schemas:
            schema.name = self.escape_special_symbol(schema.name)
            schema.doc = self.escape_special_symbol(schema.doc.strip("\n"))
            if schema.attributes:
                for attr in schema.attributes:
                    attr.name = self.escape_special_symbol(attr.name)
                    attr.doc = self.escape_special_symbol(attr.doc.strip("\n"))
                    attr.type.type_str = self.escape_special_symbol(attr.type.type_str)
            if schema.examples:
                schema.examples = schema.examples.strip("\n")
        return module

    @staticmethod
    def escape_special_symbol(name: str) -> str:
        return (
            name.replace("_", "\\_")
            .replace("*", "\\*")
            .replace("#", "\\#")
            .replace("|", "&#124;")
            .replace("<", "\\<")
            .replace(">", "\\>")
            .replace("\n", "<br />")
        )


class YamlDocEscaper(DocEscaper):
    def escape(self, module: model.ModuleDoc) -> None:
        return module


class JsonDocEscaper(DocEscaper):
    def escape(self, module: model.ModuleDoc) -> None:
        return module


factory = factory.DocEscaperFactory()
factory.register_format(doc_formats.KCLDocFormat.MARKDOWN, MarkdownDocEscaper)
factory.register_format(doc_formats.KCLDocFormat.YAML, YamlDocEscaper)
factory.register_format(doc_formats.KCLDocFormat.JSON, JsonDocEscaper)



# Copyright 2021 The KCL Authors. All rights reserved.

from .scope import (
    Scope,
    ProgramScope,
    PackageScope,
    BUILTIN_SCOPE,
    PLUGIN_SCOPE_MAPPING,
)
from .checker import ResolveProgram, CheckConfig
from .type import (
    Type,
    VOID_TYPE,
    NONE_TYPE,
    INT_TYPE,
    FLOAT_TYPE,
    STR_TYPE,
    BOOL_TYPE,
    ANY_TYPE,
    TRUE_LIT_TYPE,
    FALSE_LIT_TYPE,
    DICT_STR_ANY_TYPE,
    DICT_STR_STR_TYPE,
    DICT_ANY_ANY_TYPE,
    INT_OR_STR_TYPE,
    sup,
    assignable_to,
    is_upper_bound,
    type_to_kcl_type_annotation_str,
)
from .type_parser import parse_type_str
from .type_convension import type_convert

__all__ = [
    "Scope",
    "ProgramScope",
    "PackageScope",
    "BUILTIN_SCOPE",
    "PLUGIN_SCOPE_MAPPING",
    "ResolveProgram",
    "CheckConfig",
    "Type",
    "VOID_TYPE",
    "NONE_TYPE",
    "INT_TYPE",
    "FLOAT_TYPE",
    "STR_TYPE",
    "BOOL_TYPE",
    "ANY_TYPE",
    "TRUE_LIT_TYPE",
    "FALSE_LIT_TYPE",
    "DICT_STR_ANY_TYPE",
    "DICT_STR_STR_TYPE",
    "DICT_ANY_ANY_TYPE",
    "INT_OR_STR_TYPE",
    "sup",
    "assignable_to",
    "is_upper_bound",
    "type_to_kcl_type_annotation_str",
    "parse_type_str",
    "type_convert",
]



# Copyright 2020 The KCL Authors. All rights reserved.

from typing import cast, List

import kclvm.api.object as objpkg
import kclvm.unification.subsume as subsume

# ------------------------------------
# Type alias
# ------------------------------------

Type = objpkg.KCLBaseTypeObject

# ------------------------------------
# Type annotation str constants
# ------------------------------------

ANY_TYPE_STR = "any"
INT_TYPE_STR = "int"
FLOAT_TYPE_STR = "float"
STR_TYPE_STR = "str"
BOOL_TYPE_STR = "bool"
ITERABLE_TYPE_STR = "str|{:}|[]"
NUMBER_TYPE_STR = "int|float|bool"
NUM_OR_STR_TYPE_STR = "int|float|bool|str"
RESERVED_TYPE_IDENTIFIERS = [
    ANY_TYPE_STR,
    INT_TYPE_STR,
    FLOAT_TYPE_STR,
    STR_TYPE_STR,
    BOOL_TYPE_STR,
]

# ------------------------------------
# Type constants
# ------------------------------------

VOID_TYPE: Type = objpkg.KCLVoidTypeObject()
NONE_TYPE: Type = objpkg.KCLNoneTypeObject()
INT_TYPE: Type = objpkg.KCLIntTypeObject()
FLOAT_TYPE: Type = objpkg.KCLFloatTypeObject()
STR_TYPE: Type = objpkg.KCLStringTypeObject()
BOOL_TYPE: Type = objpkg.KCLBoolTypeObject()
ANY_TYPE: Type = objpkg.KCLAnyTypeObject()
TRUE_LIT_TYPE: Type = objpkg.KCLBoolLitTypeObject(value=True)
FALSE_LIT_TYPE: Type = objpkg.KCLBoolLitTypeObject(value=False)
NUMBER_TYPE: Type = objpkg.KCLUnionTypeObject(types=[INT_TYPE, FLOAT_TYPE, BOOL_TYPE])
NUM_OR_STR_TYPE: Type = objpkg.KCLUnionTypeObject(
    types=[INT_TYPE, FLOAT_TYPE, BOOL_TYPE, STR_TYPE]
)
LIST_ANY_TYPE: Type = objpkg.KCLListTypeObject(item_type=ANY_TYPE)
LIST_STR_TYPE: Type = objpkg.KCLListTypeObject(item_type=STR_TYPE)
DICT_ANY_ANY_TYPE: Type = objpkg.KCLDictTypeObject(
    key_type=ANY_TYPE, value_type=ANY_TYPE
)
DICT_STR_ANY_TYPE: Type = objpkg.KCLDictTypeObject(
    key_type=STR_TYPE, value_type=ANY_TYPE
)
DICT_STR_STR_TYPE: Type = objpkg.KCLDictTypeObject(
    key_type=STR_TYPE, value_type=STR_TYPE
)
INT_OR_STR_TYPE: Type = objpkg.KCLUnionTypeObject(types=[INT_TYPE, STR_TYPE])
ITERABLE_TYPE: Type = objpkg.KCLUnionTypeObject(
    types=[LIST_ANY_TYPE, DICT_ANY_ANY_TYPE, STR_TYPE]
)
LIT_TYPE_KIND_MAPPING = {
    objpkg.KCLTypeKind.StrLitKind: STR_TYPE,
    objpkg.KCLTypeKind.IntLitKind: INT_TYPE,
    objpkg.KCLTypeKind.FloatLitKind: FLOAT_TYPE,
    objpkg.KCLTypeKind.BoolLitKind: BOOL_TYPE,
}

# ------------------------------------
# Type kind constants
# ------------------------------------

NUMBER_TYPE_KINDS = [
    objpkg.KCLTypeKind.IntKind,
    objpkg.KCLTypeKind.FloatKind,
    objpkg.KCLTypeKind.IntLitKind,
    objpkg.KCLTypeKind.FloatLitKind,
    # Please note that the True/False name constant can be used to 1 + True
    objpkg.KCLTypeKind.BoolKind,
    objpkg.KCLTypeKind.BoolLitKind,
]
STR_KINDS = [
    objpkg.KCLTypeKind.StrKind,
    objpkg.KCLTypeKind.StrLitKind,
]
INT_KINDS = [
    objpkg.KCLTypeKind.IntKind,
    objpkg.KCLTypeKind.IntLitKind,
]
FLOAT_KINDS = [
    objpkg.KCLTypeKind.FloatKind,
    objpkg.KCLTypeKind.FloatLitKind,
]
BOOL_KINDS = [
    objpkg.KCLTypeKind.BoolKind,
    objpkg.KCLTypeKind.BoolLitKind,
]
ITERABLE_KINDS = [
    objpkg.KCLTypeKind.ListKind,
    objpkg.KCLTypeKind.DictKind,
    objpkg.KCLTypeKind.SchemaKind,
    objpkg.KCLTypeKind.StrKind,
    objpkg.KCLTypeKind.StrLitKind,
]
KEY_KINDS = STR_KINDS
BUILTIN_KINDS = NUMBER_TYPE_KINDS + STR_KINDS

# -----------------------
# Type functions
# -----------------------


def sup(types: List[Type]) -> Type:
    """The sup function returns the minimum supremum of all types in an array of types"""
    return typeof(types, should_remove_sub_types=True)


def typeof(types: List[Type], should_remove_sub_types: bool = False) -> Type:
    """Build a sup type from types [T1, T2, ... , Tn]"""
    assert isinstance(types, list)
    # 1. Initialize an ordered set to store the type array
    type_set = []
    # 2. Add the type array to the ordered set for sorting by the type id and de-duplication
    add_types_to_type_set(type_set, types)
    # 3. Remove sub types according to partial order relation rules e.g. sub schema types
    if should_remove_sub_types:
        type_set = remove_sub_types(type_set)
    if len(type_set) == 0:
        return ANY_TYPE
    if len(type_set) == 1:
        return type_set[0]
    type_set.sort(key=lambda t: t.type_kind())
    return objpkg.KCLUnionTypeObject(types=type_set)


def add_types_to_type_set(type_set: List[Type], types: List[Type]):
    """Add types into the type set"""
    for tpe in types or []:
        add_type_to_type_set(type_set, tpe)


def add_type_to_type_set(type_set: List[Type], tpe: Type):
    """Add a type into the type set"""
    if isinstance(tpe, objpkg.KCLUnionTypeObject):
        tpe = cast(objpkg.KCLUnionTypeObject, tpe)
        add_types_to_type_set(type_set, tpe.types)
    # Ignore 'void' types in unions
    elif not isinstance(tpe, objpkg.KCLVoidTypeObject):
        if tpe not in type_set:
            type_set.append(tpe)


def remove_sub_types(type_set: List[Type]) -> List[Type]:
    """Remove sub types from the type set"""
    remove_index_list = set()
    for i, source in enumerate(type_set):
        for j, target in enumerate(type_set):
            if i != j:
                is_subsume = subsume.type_subsume(source, target)
                if is_subsume:
                    remove_index_list.add(i)
    return [tpe for i, tpe in enumerate(type_set) if i not in remove_index_list]


def assignable_to(tpe: Type, expected_type: Type) -> bool:
    """Judge type `tpe` can be assigned the expected type"""
    if tpe.type_kind() >= objpkg.KCLTypeKind.VoidKind:
        return False
    if (
        tpe.type_kind()
        == expected_type.type_kind()
        == objpkg.KCLTypeKind.NumberMultiplierKind
    ):
        if expected_type.is_literal():
            return expected_type.type_str() == tpe.type_str()
        else:
            return True
    return subsume.type_subsume(tpe, expected_type, check_left_any=True)


def is_upper_bound(type1: Type, type2: Type) -> bool:
    """Whether `type1` is the upper bound of the `type2`"""
    return subsume.type_subsume(type2, type1)


def has_any_type(types: List[Type]):
    """Whether a type array contains any type"""
    return any([t == ANY_TYPE for t in types])


def infer_to_variable_type(tpe: Type):
    """Infer the value type to the variable type"""
    if tpe is None:
        return tpe
    # Literal type to its named type e.g., 1 -> int, "s" -> str
    if tpe.type_kind() in LIT_TYPE_KIND_MAPPING:
        return LIT_TYPE_KIND_MAPPING[tpe.type_kind()]
    # Union type e.g., 1|2|"s" -> int|str
    if tpe.type_kind() == objpkg.KCLTypeKind.UnionKind:
        return sup([infer_to_variable_type(t) for t in tpe.types])
    # List type e.g., [1|2] -> [int]
    if tpe.type_kind() == objpkg.KCLTypeKind.ListKind:
        tpe.item_type = infer_to_variable_type(tpe.item_type)
    # Dict type e.g., {str:1|2} -> {str:int}
    if tpe.type_kind() == objpkg.KCLTypeKind.DictKind:
        tpe.key_type = infer_to_variable_type(tpe.key_type)
        tpe.value_type = infer_to_variable_type(tpe.value_type)
    # None/Undefined type to any type e.g., None -> any
    if tpe == NONE_TYPE:
        return ANY_TYPE
    return tpe


def literal_union_type_to_variable_type(tpe: Type):
    """Convert the literal union type to its variable type
    e.g., 1|2 -> int, 's'|'ss' -> str.
    """
    if tpe.type_kind() == objpkg.KCLTypeKind.UnionKind:
        return infer_to_variable_type(tpe)
    return tpe


def is_kind_type_or_kind_union_type(tpe: Type, kind_list: List[int]):
    """Judge a type kind in the type kind list or the union
    type kinds are all in the type kind.
    """
    result = False
    if tpe.type_kind() == objpkg.KCLTypeKind.UnionKind:
        result = all([t.type_kind() in kind_list for t in tpe.types])
    elif tpe.type_kind() in kind_list:
        result = True
    return result


def type_to_kcl_type_annotation_str(tpe: Type) -> str:
    """Convert type to a kcl type annotation string"""
    if not tpe or not isinstance(tpe, Type):
        raise ValueError(f"Parameter type must be a valid type, got {tpe}")
    if isinstance(tpe, objpkg.KCLUnionTypeObject):
        return "|".join([type_to_kcl_type_annotation_str(t) for t in tpe.types])
    elif isinstance(tpe, objpkg.KCLIntTypeObject):
        return "int"
    elif isinstance(tpe, objpkg.KCLFloatTypeObject):
        return "float"
    elif isinstance(tpe, objpkg.KCLStringTypeObject):
        return "str"
    elif isinstance(tpe, objpkg.KCLBoolTypeObject):
        return "bool"
    elif isinstance(tpe, objpkg.KCLAnyTypeObject):
        return "any"
    elif isinstance(tpe, objpkg.KCLStringLitTypeObject):
        return '"{}"'.format(tpe.value.replace('"', '\\"'))
    elif isinstance(
        tpe,
        (
            objpkg.KCLIntLitTypeObject,
            objpkg.KCLFloatLitTypeObject,
            objpkg.KCLBoolLitTypeObject,
        ),
    ):
        return str(tpe.value)
    elif isinstance(tpe, objpkg.KCLListTypeObject):
        return "[" + type_to_kcl_type_annotation_str(tpe.item_type) + "]"
    elif isinstance(tpe, objpkg.KCLDictTypeObject):
        return (
            "{"
            + type_to_kcl_type_annotation_str(tpe.key_type)
            + ":"
            + type_to_kcl_type_annotation_str(tpe.value_type)
            + "}"
        )
    elif isinstance(tpe, objpkg.KCLSchemaDefTypeObject):
        return tpe.schema_type.type_str_with_pkgpath()
    elif isinstance(tpe, objpkg.KCLSchemaTypeObject):
        return tpe.type_str_with_pkgpath()
    elif isinstance(tpe, objpkg.KCLNumberMultiplierTypeObject):
        return (
            f"{tpe.raw_value}{tpe.binary_suffix}"
            if tpe.is_literal()
            else "units.NumberMultiplier"
        )
    return ""



# Copyright 2020 The KCL Authors. All rights reserved.

from typing import Union

import kclvm.api.object as objpkg

# --------------------------------------
# Numerical Constants
# Usage:
#     import units
#     memory = 1024 * units.Mi  # 1024Mi
# --------------------------------------

KMANGLED_n = 1e-09
KMANGLED_u = 1e-06
KMANGLED_m = 0.001
KMANGLED_k = 1_000
KMANGLED_K = 1_000
KMANGLED_M = 1_000_000
KMANGLED_G = 1_000_000_000
KMANGLED_T = 1_000_000_000_000
KMANGLED_P = 1_000_000_000_000_000
KMANGLED_Ki = 1024
KMANGLED_Mi = 1024**2
KMANGLED_Gi = 1024**3
KMANGLED_Ti = 1024**4
KMANGLED_Pi = 1024**5

UNIT_MAPPING = {
    "n": KMANGLED_n,
    "u": KMANGLED_u,
    "m": KMANGLED_m,
    "k": KMANGLED_k,
    "K": KMANGLED_K,
    "M": KMANGLED_M,
    "G": KMANGLED_G,
    "T": KMANGLED_T,
    "P": KMANGLED_P,
    "Ki": KMANGLED_Ki,
    "Mi": KMANGLED_Mi,
    "Gi": KMANGLED_Gi,
    "Ti": KMANGLED_Ti,
    "Pi": KMANGLED_Pi,
}

# --------------------------------------
# Numerical Multiplier Type
# Usage:
#     import units
#     memory: units.NumberMultiplier = 1M
# --------------------------------------

KMANGLED_NumberMultiplier = objpkg.KCLNumberMultiplierTypeObject()

# ------------------------------------------
# Unit ToString Methods
# Usage:
#     import units
#     disk = units.to_Ki(1024)  # "1Ki"
# Input:
#     num: int
# Returns:
#     int
# Raises:
#     ValueError on invalid or unknown input
# ------------------------------------------


def KMANGLED_to_n(num: int) -> str:
    """Int literal to string with `n` suffix"""
    return to_unit(num, "n")


def KMANGLED_to_u(num: int) -> str:
    """Int literal to string with `u` suffix"""
    return to_unit(num, "u")


def KMANGLED_to_m(num: int) -> str:
    """Int literal to string with `m` suffix"""
    return to_unit(num, "m")


def KMANGLED_to_K(num: int) -> str:
    """Int literal to string with `K` suffix"""
    return to_unit(num, "K")


def KMANGLED_to_M(num: int) -> str:
    """Int literal to string with `M` suffix"""
    return to_unit(num, "M")


def KMANGLED_to_G(num: int) -> str:
    """Int literal to string with `G` suffix"""
    return to_unit(num, "G")


def KMANGLED_to_T(num: int) -> str:
    """Int literal to string with `T` suffix"""
    return to_unit(num, "T")


def KMANGLED_to_P(num: int) -> str:
    """Int literal to string with `P` suffix"""
    return to_unit(num, "P")


def KMANGLED_to_Ki(num: int) -> str:
    """Int literal to string with `Ki` suffix"""
    return to_unit(num, "Ki")


def KMANGLED_to_Mi(num: int) -> str:
    """Int literal to string with `Mi` suffix"""
    return to_unit(num, "Mi")


def KMANGLED_to_Gi(num: int) -> str:
    """Int literal to string with `Gi` suffix"""
    return to_unit(num, "Gi")


def KMANGLED_to_Ti(num: int) -> str:
    """Int literal to string with `Ti` suffix"""
    return to_unit(num, "Ti")


def KMANGLED_to_Pi(num: int) -> str:
    """Int literal to string with `Pi` suffix"""
    return to_unit(num, "Pi")


def to_unit(num: Union[int, float], suffix: str) -> str:
    """Connect numbers and suffixes"""
    if not isinstance(num, (int, float)):
        raise ValueError("Unsupported number type: {}".format(type(num)))
    if not suffix or not isinstance(suffix, str) or suffix not in list(UNIT_MAPPING):
        raise ValueError("Unsupported unit suffix: {}".format(suffix))
    return str(int(num // UNIT_MAPPING[suffix])) + suffix



from .merge import MergeAST, MergeASTList, MergeASTToVertex, MergeStrategy
from .vertex import Vertex
from .unifier import Unifier
from .subsume import value_subsume, type_subsume

__all__ = [
    "MergeASTList",
    "MergeAST",
    "MergeASTToVertex",
    "MergeStrategy",
    "Vertex",
    "Unifier",
    "value_subsume",
    "type_subsume",
]



from typing import Callable, Optional, Dict
from functools import wraps
import inspect

import kclvm.kcl.info as kcl_info

from .plugin import get_plugin_names, get_plugin
from kclvm.api.object import KCLBuiltinFunctionObject


PLUGIN_MODULE_NAME = "kcl_plugin."
STANDARD_SYSTEM_MODULE_LOCATION = "kclvm_plugin"


def kcl_plugin(func):
    @wraps(func)
    def decorated(*args, **kwargs):
        return func(*args, **kwargs)

    return decorated


def new_plugin_function(
    name: str, func: Callable
) -> Optional[KCLBuiltinFunctionObject]:
    """New a plugin function object using a native plugin function"""
    if not func or not name:
        return None
    return KCLBuiltinFunctionObject(name=name, function=func)


def get_plugin_func_objects(plugin_name: str) -> Dict[str, KCLBuiltinFunctionObject]:
    """Get all plugin function objects from a plugin named 'plugin_name'"""
    if (
        not plugin_name
        or plugin_name.replace(PLUGIN_MODULE_NAME, "") not in get_plugin_names()
    ):
        return {}
    module = get_plugin(plugin_name)
    members = inspect.getmembers(module)
    result = {
        kcl_info.demangle(func_name): new_plugin_function(
            kcl_info.demangle(func_name), func
        )
        for func_name, func in members
        if kcl_info.ismangled(func_name)
    }
    return result



# Copyright 2021 The KCL Authors. All rights reserved.

MANGLE_PREFIX = "KMANGLED_"
TAGGING_PREFIX = "KTAG_"


def mangle(name):
    """Mangle a name"""
    dot = name.rfind(".")
    if dot == -1:
        mangled_name = MANGLE_PREFIX + name
    else:
        mangled_name = mangle(name[:dot]) + "." + mangle(name[dot + 1 :])
    return mangled_name


def demangle(name):
    """Demangle a name if it is mangled"""
    demangled_name = ""
    if ismangled(name):
        dot = name.rfind(".")
        if dot == -1:
            assert len(name) > len(
                MANGLE_PREFIX
            ), "Internal Error: Demangling failure. Please report a bug to us."
            demangled_name += name[len(MANGLE_PREFIX) :]
        else:
            demangled_name = demangle(name[:dot]) + "." + demangle(name[dot + 1 :])
    else:
        demangled_name = name
    return demangled_name


def ismangled(name):
    """Check if a name is mangled"""
    if name.startswith(MANGLE_PREFIX):
        return True
    return False


def tagging(tag, name=None):
    """tagging a name"""
    return TAGGING_PREFIX + tag + "_" + name


def detagging(tag, name=None):
    """Detagging a name if it is tagged"""
    if istagged(name):
        return name[len(TAGGING_PREFIX) + len(tag) + 1 :]
    return name


def istagged(name):
    """Check if a name is tagged"""
    if name.startswith(TAGGING_PREFIX):
        return True
    return False


def isprivate_field(name):
    return name.startswith("_")
